import abc
import re


class Rule(abc.ABC):
    @abc.abstractmethod
    def matches(self, line: str) -> bool: ...

    @abc.abstractmethod
    def __hash__(self) -> int: ...


class RegexRule(Rule):
    def __init__(self, pattern: str, flags: "re._FlagsType"):
        self.pattern = re.compile(pattern, flags)

    def matches(self, line: str) -> bool:
        return self.pattern.search(line) != None

    def __hash__(self) -> int:
        return hash(self.pattern)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, RegexRule):
            return False
        return self.pattern == other.pattern


class LiteralRule(Rule):
    def __init__(self, pattern: str, ignorecase: bool):
        self.ignorecase = ignorecase
        if ignorecase:
            self.pattern = pattern.casefold()
        else:
            self.pattern = pattern

    def matches(self, line: str) -> bool:
        if self.ignorecase:
            line = line.casefold()
        return self.pattern in line

    def __hash__(self) -> int:
        return hash((self.ignorecase, self.pattern))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, LiteralRule):
            return False
        return self.pattern == other.pattern and self.ignorecase == other.ignorecase



import threading
from collections import deque
from typing import Generator

from minimux.rules import Rule


class Buffer:
    def __init__(
        self,
        maxcols: int,
        maxrows: int,
        rules: dict[Rule, int] | None = None,
    ):
        self.buf: deque[tuple[str, int]] = deque(maxlen=maxrows)
        self.maxcols = maxcols
        self.rules = rules if rules is not None else {}
        self.lock = threading.Lock()

    def push(self, data: str):
        with self.lock:
            if self.maxcols == 0:
                return

            attr = 0
            for rule, a in self.rules.items():
                if rule.matches(data):
                    attr = a
                    break

            for line in data.splitlines(keepends=False):
                while len(line) > 0:
                    b, line = line[: self.maxcols], line[self.maxcols :]
                    self.buf.append((b, attr))

    def resize(
        self,
        *,
        maxcols: int | None = None,
        maxrows: int | None = None,
    ):
        with self.lock:
            if maxcols is not None:
                self.maxcols = maxcols
            if maxrows is not None:
                self.maxrows = maxrows

            buf: deque[tuple[str, int]] = deque(maxlen=maxrows)
            for line in self.buf:
                buf.append((line[0][: self.maxcols], line[1]))
            self.buf = buf

    def __iter__(self) -> Generator[tuple[str, int], None, None]:
        with self.lock:
            yield from self.buf



import configparser
import curses
import re
import shlex
from dataclasses import dataclass
from typing import TextIO

import minimux.utils as utils
from minimux.colour import ColourManager
from minimux.rules import LiteralRule, RegexRule, Rule


@dataclass
class Attr:
    fg: str | int | None = None
    bg: str | int | None = None
    blink: bool | None = None
    bold: bool | None = None
    dim: bool | None = None
    reverse: bool | None = None
    standout: bool | None = None
    underline: bool | None = None

    def __or__(self, other: object) -> "Attr":
        if not isinstance(other, Attr):
            raise TypeError
        return Attr(
            fg=utils.combine(self.fg, other.fg),
            bg=utils.combine(self.bg, other.bg),
            blink=utils.combine(self.blink, other.blink),
            bold=utils.combine(self.bold, other.bold),
            dim=utils.combine(self.dim, other.dim),
            reverse=utils.combine(self.reverse, other.reverse),
            standout=utils.combine(self.standout, other.standout),
            underline=utils.combine(self.underline, other.underline),
        )

    def __call__(self, colour_manager: ColourManager):
        attr = colour_manager.make_pair(self.fg, self.bg)
        if self.blink:
            attr |= curses.A_BLINK
        if self.bold:
            attr |= curses.A_BOLD
        if self.dim:
            attr |= curses.A_DIM
        if self.reverse:
            attr |= curses.A_REVERSE
        if self.standout:
            attr |= curses.A_STANDOUT
        if self.underline:
            attr |= curses.A_UNDERLINE
        return attr


@dataclass
class Element:
    name: str
    attr: Attr
    weight: int


@dataclass
class Command(Element):
    command: list[str]
    title: str | None
    rules: dict[Rule, Attr]
    shell: bool
    input: str | None
    padding: tuple[int, int, int, int]
    title_attr: Attr


@dataclass
class Panel(Element):
    vertical: bool
    children: list[Element]


class MiniMuxConfigParser(configparser.ConfigParser):
    def __init__(self):
        super().__init__(
            delimiters=["="],
            comment_prefixes=["#"],
            default_section="default",
        )

    def create_panels(
        self,
        section: "configparser.SectionProxy",
        prefix: str,
        default_attr: Attr,
    ) -> Element:
        if "command" in section:
            return self.parse_command(section, prefix, default_attr)
        elif "panels" in section:
            return self.parse_panel(section, prefix, default_attr)
        else:
            raise ValueError("command or panels must be specified")

    def parse_command(
        self,
        section: "configparser.SectionProxy",
        prefix: str,
        default_attr: Attr,
    ) -> Command:
        title = section.get("title", None)
        command = section["command"]
        attr = default_attr | self.parse_attrs(section)
        title_attr = attr
        if s := section.get("title_attr", None):
            title_attr = attr | self.parse_attrs(self[s])
        rules = self.parse_rules(self.aslist(section.get("rules", "")), attr)
        weight = section.getint("weight", 1)
        shell = section.getboolean("shell", False)
        input = section.get("input", None)
        padding = self.aspadding(section.get("padding", None))

        return Command(
            prefix + ":" + section.name,
            attr,
            weight,
            shlex.split(command),
            title,
            rules,
            shell,
            input,
            padding,
            title_attr,
        )

    def parse_panel(
        self,
        section: "configparser.SectionProxy",
        prefix: str,
        default_attr: Attr,
    ) -> Panel:
        prefix += ":" + section.name
        vertical = section.getboolean("vertical", False)
        attr = default_attr | self.parse_attrs(section)
        children: list[Element] = []
        weight = section.getint("weight", 1)
        for subsection in self.aslist(section["panels"]):
            child = self.create_panels(self[subsection], prefix, attr)
            children.append(child)

        return Panel(prefix, attr, weight, vertical, children)

    def parse_rules(
        self, rule_names: list[str], default_attr: Attr
    ) -> dict[Rule, Attr]:
        rules: dict[Rule, Attr] = {}
        for rule_name in rule_names:
            rule, attr = self.parse_rule(rule_name, default_attr)
            rules[rule] = attr
        return rules

    def parse_rule(self, rule_name: str, default_attr: Attr) -> tuple[Rule, Attr]:
        rule = self[rule_name]
        if "regex" in rule:
            return self.parse_regex_rule(rule, default_attr)
        elif "literal" in rule:
            return self.parse_literal_rule(rule, default_attr)
        else:
            raise ValueError("invalid rule: must contain 'regex' or 'literal' option")

    def parse_regex_rule(
        self,
        section: "configparser.SectionProxy",
        default_attr: Attr,
    ) -> tuple[Rule, Attr]:
        pattern = section["regex"]
        flags = re.NOFLAG
        if section.getboolean("ascii", False):
            flags |= re.ASCII
        if section.getboolean("ignorecase", False):
            flags |= re.IGNORECASE
        if section.getboolean("locale", False):
            flags |= re.LOCALE
        if section.getboolean("multiline", False):
            flags |= re.MULTILINE
        if section.getboolean("dotall", False):
            flags |= re.DOTALL
        if section.getboolean("verbose", False):
            flags |= re.VERBOSE

        rule = RegexRule(pattern, flags)
        attr = default_attr | self.parse_attrs(section)
        return rule, attr

    def parse_literal_rule(
        self,
        section: "configparser.SectionProxy",
        default_attr: Attr,
    ) -> tuple[Rule, Attr]:
        pattern = section["literal"]
        ignorecase = section.getboolean("ignorecase", False)

        rule = LiteralRule(pattern, ignorecase)
        attr = default_attr | self.parse_attrs(section)
        return rule, attr

    def parse_attrs(self, section: "configparser.SectionProxy") -> Attr:
        attr = Attr()
        attr.fg = section.get("fg", None)
        attr.bg = section.get("bg", None)
        attr.blink = section.getboolean("blink", None)
        attr.bold = section.getboolean("bold", None)
        attr.dim = section.getboolean("dim", None)
        attr.reverse = section.getboolean("reverse", None)
        attr.standout = section.getboolean("standout", None)
        attr.underline = section.getboolean("underline", None)
        return attr

    def aslist(self, value: str):
        return [v.strip() for v in value.split(",") if v.strip()]

    def aspadding(self, value: str | None) -> tuple[int, int, int, int]:
        if value is None:
            return (0, 0, 0, 0)
        ps = [int(v) for v in value.split(",")]
        if len(ps) == 1:
            return (ps[0], ps[0], ps[0], ps[0])
        elif len(ps) == 2:
            return (ps[1], ps[0], ps[1], ps[0])
        elif len(ps) == 4:
            return (ps[0], ps[1], ps[2], ps[3])
        raise ValueError("expected p or px,py or pt,pr,pb,pl")


@dataclass
class Config:
    title: str | None
    content: Element
    sep_attr: Attr
    title_attr: Attr

    @classmethod
    def from_parser(cls, parser: MiniMuxConfigParser) -> "Config":
        main = parser["main"]
        title = main.pop("title", None)
        content = parser.create_panels(main, "", Attr())
        base_attr = parser.parse_attrs(main)
        sep_attrs = base_attr
        if "seperator" in parser:
            sep_attrs = base_attr | parser.parse_attrs(parser["seperator"])
        title_attrs = base_attr
        if "title" in parser:
            title_attrs = base_attr | parser.parse_attrs(parser["title"])

        return cls(title, content, sep_attrs, title_attrs)

    @classmethod
    def from_file(cls, f: TextIO) -> "Config":
        parser = MiniMuxConfigParser()
        parser.read_file(f)
        return cls.from_parser(parser)



import curses


def _scale(v: int) -> int:
    return int(v * 1000 / 256)


class ColourManager:
    def __init__(self):
        self.colours = {
            "black": curses.COLOR_BLACK,
            "blue": curses.COLOR_BLUE,
            "cyan": curses.COLOR_CYAN,
            "green": curses.COLOR_GREEN,
            "magenta": curses.COLOR_MAGENTA,
            "red": curses.COLOR_RED,
            "white": curses.COLOR_WHITE,
            "yellow": curses.COLOR_YELLOW,
        }
        self.next_colour = 8
        self.colour_pairs = []

    def make_pair(self, fg: str | int | None, bg: str | int | None) -> int:
        if fg is None and bg is None:
            return 0

        # colour string to curses colour
        f = self.parse_colour(fg)
        b = self.parse_colour(bg)

        try:
            # try to find cached value
            n = self.colour_pairs.index((f, b))
        except ValueError:
            # create new colour pair and cache
            n = len(self.colour_pairs)
            curses.init_pair(n + 1, f, b)
            self.colour_pairs.append((f, b))

        return curses.color_pair(n + 1)

    def parse_colour(self, colour: str | int | None) -> int:
        # int values are already parsed
        if isinstance(colour, int):
            return colour

        # empty string or none signals default
        if not colour:
            return -1

        # check for cached colours
        if colour in self.colours:
            return self.colours[colour]

        # try to parse
        if colour.startswith("#"):
            return self.parse_hex(colour)
        elif colour.startswith("rgb("):
            return self.parse_rgb(colour)
        else:
            raise ValueError("invalid colour: " + colour)

    def parse_hex(self, hex_code: str) -> int:
        res = self.next_colour
        self.next_colour += 1
        curses.init_color(
            res,
            _scale(int(hex_code[1:3], 16)),
            _scale(int(hex_code[3:5], 16)),
            _scale(int(hex_code[5:7], 16)),
        )
        return res

    def parse_rgb(self, rgb: str) -> int:
        res = self.next_colour
        self.next_colour += 1
        r, g, b = rgb[4:-1].split(",")
        curses.init_color(
            res,
            _scale(int(r)),
            _scale(int(g)),
            _scale(int(b)),
        )
        return res



from pathlib import Path

import click

from minimux import MiniMux, __version__
from minimux.config import Config, MiniMuxConfigParser


@click.command("minimux")
@click.version_option(__version__, "--version", "-v")
@click.argument(
    "config_file",
    type=click.Path(
        exists=True,
        dir_okay=False,
        path_type=Path,
    ),
    default=Path("./minimux.ini"),
)
def main(config_file: Path):
    # load config
    parser = MiniMuxConfigParser()
    with open(config_file) as f:
        parser.read_file(f)
    config = Config.from_parser(parser)

    # run
    try:
        minimux = MiniMux(config)
        minimux.run()
    except Exception as e:
        exit(str(e))


if __name__ == "__main__":
    main()



import curses
import threading

import minimux.utils as utils
from minimux.colour import ColourManager
from minimux.config import Command, Config, Element, Panel
from minimux.runner import Runner

__version__ = "1.3.0"
__version_info__ = (1, 3, 0)
__author__ = "Dominic Price"


class MiniMux:
    def __init__(
        self,
        config: Config,
    ):
        self.cm = ColourManager()
        self.config = config
        self.lock = threading.Lock()
        self.runners: dict[str, Runner] = {}

    def run(self):
        """The main entrypoint"""
        curses.wrapper(self._run)

    def _run(self, stdscr: "curses._CursesWindow"):
        curses.curs_set(False)
        curses.use_default_colors()

        self.runners = self.get_runners(self.config.content)
        for runner in self.runners.values():
            runner.start()
        self.init(stdscr)

        try:
            # handle user input
            while True:
                if stdscr.getch() == curses.KEY_RESIZE:
                    self.init(stdscr)
        except KeyboardInterrupt:
            # keyboard interrupts are a normal way to exit
            pass
        finally:
            # kill all runners and wait for them to terminate
            # before exiting curses mode to avoid leaving the
            # terminal in a bad state
            rows, cols = stdscr.getmaxyx()
            wait_fns = [runner.terminate() for runner in self.runners.values()]
            stdscr.clear()
            stdscr.move(rows // 2, 0)
            stdscr.addstr("Terminating...".center(cols))
            stdscr.refresh()
            for wait_fn in wait_fns:
                wait_fn()

    def get_runners(self, content: Element) -> dict[str, Runner]:
        if isinstance(content, Panel):
            res: dict[str, Runner] = {}
            for child in content.children:
                res.update(self.get_runners(child))
            return res
        elif isinstance(content, Command):
            return {content.name: Runner(content, self.lock, self.cm)}
        else:
            raise TypeError

    def init(self, stdscr: "curses._CursesWindow"):
        rows, cols = stdscr.getmaxyx()
        stdscr.clear()

        start_row = 0
        if self.config.title:
            stdscr.move(0, 0)
            stdscr.addstr(
                self.config.title.center(cols),
                self.config.title_attr(self.cm),
            )
            self.hsep(stdscr, 1, 0, cols)
            start_row = 2

        self.init_content(
            stdscr,
            self.config.content,
            (start_row, rows),
            (0, cols),
        )
        stdscr.refresh()

    def init_content(
        self,
        stdscr: "curses._CursesWindow",
        content: Element,
        range_y: tuple[int, int],
        range_x: tuple[int, int],
    ):
        """Recursively draw the static components for content and
        initialise the runners for any commands"""
        if isinstance(content, Panel):
            self.init_panel(stdscr, content, range_y, range_x)
        elif isinstance(content, Command):
            self.init_command(stdscr, content, range_y, range_x)
        else:
            raise TypeError

    def init_panel(
        self,
        stdscr: "curses._CursesWindow",
        panel: Panel,
        range_y: tuple[int, int],
        range_x: tuple[int, int],
    ):
        """Recursively draw the static components for a panel and
        initialise the runners from any commands"""
        if len(panel.children) == 0:
            return []

        if panel.vertical:
            o = range_y[0]
            subh = (range_y[1] - range_y[0]) // sum(c.weight for c in panel.children)
            i = 0
            for child in panel.children:
                subrange_y = (o + i * subh, o + (i + child.weight) * subh)
                if i == len(panel.children) - 1:
                    subrange_y = (subrange_y[0], range_y[1])
                if i != 0:
                    self.hsep(
                        stdscr,
                        subrange_y[0],
                        range_x[0],
                        range_x[1] - range_x[0],
                    )
                    subrange_y = (subrange_y[0] + 1, subrange_y[1])
                i += child.weight
                self.init_content(stdscr, child, subrange_y, range_x)
        else:
            i = 0
            o = range_x[0]
            subw = (range_x[1] - range_x[0]) // sum(c.weight for c in panel.children)
            for child in panel.children:
                subrange_x = (o + i * subw, o + (i + child.weight) * subw)
                subrange_y = range_y
                if i == len(panel.children) - 1:
                    subrange_x = (subrange_x[0], range_x[1])
                if i != 0:
                    self.vsep(
                        stdscr,
                        range_y[0],
                        subrange_x[0],
                        range_y[1] - range_y[0],
                    )
                    subrange_x = (subrange_x[0] + 1, subrange_x[1])
                self.init_content(stdscr, child, subrange_y, subrange_x)
                i += child.weight

    def init_command(
        self,
        stdscr: "curses._CursesWindow",
        command: Command,
        range_y: tuple[int, int],
        range_x: tuple[int, int],
    ):
        """Draw the static components for a command initialise the
        associated runner"""
        if command.title is not None:
            stdscr.move(range_y[0], range_x[0])
            stdscr.addstr(" " * (range_x[1] - range_x[0]), command.attr(self.cm))
            self.center(
                stdscr,
                command.title,
                range_y[0],
                range_x[0],
                range_x[1] - range_x[0],
                command.title_attr(self.cm),
            )
            range_y = (range_y[0] + 1, range_y[1])
        self.runners[command.name].init(
            stdscr,
            (
                range_y[1] - range_y[0],
                range_x[1] - range_x[0],
                range_y[0],
                range_x[0],
            ),
        )

    def center(
        self,
        win: "curses._CursesWindow",
        s: str,
        y: int,
        x: int,
        n: int,
        attr: int,
    ):
        pad_left = (n - len(s)) // 2
        win.move(y, x + pad_left)
        win.addstr(s, attr)

    def hsep(self, stdscr: "curses._CursesWindow", y: int, x: int, n: int):
        """Draw a horizontal seperator line, combining with existing
        separators to form tees and crosses"""
        attr = self.config.sep_attr(self.cm)
        if x > 0 and utils.compare_char(stdscr.inch(y, x - 1), curses.ACS_SBSB):
            x -= 1
            n += 1
        for i in range(n):
            stdscr.move(y, x + i)
            cross = utils.compare_char(stdscr.inch(y, x + i), curses.ACS_SBSB)
            if cross:
                if i == 0:
                    stdscr.addch(curses.ACS_SSSB, attr)
                elif i == n - 1:
                    stdscr.addch(curses.ACS_SBSS, attr)
                else:
                    stdscr.addch(curses.ACS_SSSS, attr)
            else:
                stdscr.addch(curses.ACS_BSBS, attr)
        stdscr.noutrefresh()

    def vsep(self, stdscr: "curses._CursesWindow", y: int, x: int, n: int):
        """Draw a vertical seperator line, combining with existing
        separators to form tees and crosses"""
        attr = self.config.sep_attr(self.cm)
        if y > 0 and utils.compare_char(stdscr.inch(y - 1, x), curses.ACS_BSBS):
            y -= 1
            n += 1
        for i in range(n):
            cross = utils.compare_char(stdscr.inch(y + i, x), curses.ACS_BSBS)
            stdscr.move(y + i, x)
            if cross:
                if i == 0:
                    stdscr.addch(curses.ACS_BSSS, attr)
                elif i == n - 1:
                    stdscr.addch(curses.ACS_SSBS, attr)
                else:
                    stdscr.addch(curses.ACS_SSSS, attr)
            else:
                stdscr.addch(curses.ACS_SBSB, attr)
        stdscr.noutrefresh()



import atexit
import curses
import subprocess
import threading
from typing import Any, Callable, TypeAlias

from minimux.buffer import Buffer
from minimux.colour import ColourManager
from minimux.config import Command

WindowBounds: TypeAlias = tuple[int, int, int, int]


class Runner:
    def __init__(
        self,
        command: Command,
        lock: threading.Lock,
        colour_manager: ColourManager,
    ):
        self.command = command
        self.lock = lock
        self.win: "curses._CursesWindow | None" = None
        self.proc: subprocess.Popen[str] | None = None
        self.bkgd = command.attr(colour_manager)

        rules = {r: a(colour_manager) for r, a in command.rules.items()}
        self.buf = Buffer(0, 0, rules)

    def init(self, stdscr: "curses._CursesWindow", bounds: WindowBounds):
        with self.lock:
            if self.win is not None:
                del self.win
            self.win = stdscr.subwin(*bounds)
            self.win.bkgdset(" ", self.bkgd)
            pt, pr, pb, pl = self.command.padding
            self.buf.resize(
                maxrows=bounds[0] - pt - pb,
                maxcols=bounds[1] - pl - pr,
            )
        self.flush()

    def start(self):
        t = threading.Thread(target=self.run, daemon=True)
        t.start()

    def run(self):
        try:
            self.proc = subprocess.Popen(
                self.command.command,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                stdin=subprocess.PIPE,
                shell=self.command.shell,
                text=True,
            )

            # ensure the program is terminated at exit
            atexit.register(self.proc.kill)

            # ensure we are connected to stdin/stdout
            assert self.proc.stdin is not None
            assert self.proc.stdout is not None
        except Exception as e:
            self.buf.push("error: failed to start process: " + str(e))
            self.flush()
            return

        if self.command.input is not None:
            self.proc.stdin.write(self.command.input + "\n")
            self.proc.stdin.flush()
        self.proc.stdin.close()

        while self.proc.poll() is None:
            stdout = self.proc.stdout.readline().rstrip()
            self.buf.push(stdout)
            self.flush()

        for line in self.proc.stdout.readlines():
            self.buf.push(line.rstrip())

        self.buf.push(f"** Process exited with status code {self.proc.poll()} **")
        self.flush()

    def terminate(self) -> Callable[[], Any]:
        with self.lock:
            self.win = None
        if self.proc != None:
            self.proc.kill()
            return self.proc.wait
        return lambda: None

    def flush(self):
        with self.lock:
            if self.win is None:
                return
            self.win.clear()
            for i, (line, attr) in enumerate(self.buf):
                self.win.move(
                    i + self.command.padding[0],
                    self.command.padding[3],
                )
                self.win.addstr(line, attr)
            self.win.refresh()
            curses.doupdate()



import curses
from typing import TypeVar

T = TypeVar("T")


def combine(t1: T, t2: T) -> T:
    if t2 is None:
        return t1
    return t2


def compare_char(a: int, b: int) -> bool:
    return (a & curses.A_CHARTEXT) == (b & curses.A_CHARTEXT)
